/*
 * Copyright (c) 1983 Eric P. Allman
 * Copyright (c) 1988 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted provided
 * that: (1) source distributions retain this entire copyright notice and
 * comment, and (2) distributions including binaries display the following
 * acknowledgement:  ``This product includes software developed by the
 * University of California, Berkeley and its contributors'' in the
 * documentation or other materials provided with the distribution and in
 * all advertising materials mentioning features or use of this software.
 * Neither the name of the University nor the names of its contributors may
 * be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */


/*
** MAKECONNECTION -- Establish SMTP connection to remote host
** ----------------------------------------------------------
**
**	Returns:
**		Status code indicating success or failure.
**
**	Outputs:
**		Sets outfile and infile to the output and input channel.
*/

static int makeconnection(
  char *host,				/* host name to connect to */
  int *outfile,				/* smtp output channel */
  int *infile 				/* smtp input channel */
)
{
	struct hostent *hp;
	struct servent *sp;
	struct sockaddr_in sin;
	static char hostname[MAXHOSTNAME+1];
	struct in_addr inaddr[MAXADDRS];
	int naddrs;
	u_long addr;
	int sock;
	register int i;

/*
 * Reset state.
 */

        memset(&sin, 0, sizeof(sin));

	errno = 0;
	h_errno = 0;

	if (host == NULL || host[0] == '\0')
		host = "localhost";

	addr = inet_addr(host);

	{
		hp = gethostbyname(host);
		if (hp == NULL)
		{
			/* cannot contact nameserver, force retry */
			if (errno == ETIMEDOUT || errno == ECONNREFUSED)
				h_errno = TRY_AGAIN;

			/* nameserver could not resolve name properly */
			if (h_errno == TRY_AGAIN)
				return(EX_TEMPFAIL);

			/* no address found by nameserver */
			return(EX_NOHOST);
		}
		host = hp->h_name;

		for (i = 0; i < MAXADDRS && hp->h_addr_list[i]; i++)
			inaddr[i] = incopy(hp->h_addr_list[i]);
		naddrs = i;
	}

	(void) strncpy(hostname, host, MAXHOSTNAME);

/*
 * Try to make connection to each of the addresses in turn.
 */
	sp = getservbyname("smtp", "tcp");
	if (sp == NULL)
		return(EX_OSERR);

	for (i = 0; i < naddrs; i++)
	{
		sin.sin_family = AF_INET;
		sin.sin_port = sp->s_port;
		sin.sin_addr = inaddr[i];

		sock = socket(AF_INET, SOCK_STREAM, 0);
		if (sock < 0)
			return(EX_TEMPFAIL);

		if (verbose >= 2 || debug)
		{
			printf("connecting to %s port %d\n",
				inet_ntoa(sin.sin_addr), sin.sin_port);
	        }

		if (setjmp(Timeout) != 0)
		{
			(void) close(sock);
			errno = ETIMEDOUT;
			continue;
		}

		(void) signal(SIGALRM, timer);
		(void) alarm((unsigned)ConnTimeout);

		if (connect(sock, (struct sockaddr *)&sin, sizeof(sin)) < 0)
		{
			int err = errno;
			(void) alarm((unsigned)0);
			(void) close(sock);
			errno = err;
			if (errno == EINTR)
				errno = ETIMEDOUT;
			if (errno == ECONNREFUSED)
				return(EX_TEMPFAIL);
			continue;
		}
		(void) alarm((unsigned)0);

		*outfile = sock;
		*infile  = dup(sock);

		errno = 0;
		h_errno = 0;
		return(EX_OK);
	}

	return(EX_TEMPFAIL);
}
